from pydantic import BaseModel
from typing import Optional


class WorkerConfig(BaseModel):
    CELERY_BROKER_URL: str
    CELERY_BACKEND_URL: str
    LOGLEVEL: Optional[str] = "info"
    CONCURRENCY: Optional[int] = 1
    PREFETCH_MULTIPLIER: Optional[int] = 1
    RESULTS_EXPIRES: Optional[int] = 3600
    BROKER_TRANSPORT_OPTIONS: Optional[dict] = {
        "max_retries": 5,
        "interval_start": 0,
        "interval_step": 0.1,
        "interval_max": 0.5,
    }
    REDIS_RETRY_ON_TIMEOUT: bool = True
    REDIS_SOCKET_KEEPALIVE: bool = True



from flask import Flask, Blueprint
from oshepherd.api.config import ApiConfig
from oshepherd.api.generate.routes import generate_blueprint
from oshepherd.api.chat.routes import chat_blueprint
from oshepherd.api.embeddings.routes import embeddings_blueprint
from oshepherd.worker.app import create_celery_app_for_flask


def start_flask_app(config: ApiConfig):
    app = Flask(config.FLASK_PROJECT_NAME)
    app.config["FLASK_RUN_PORT"] = config.FLASK_RUN_PORT
    app.config["FLASK_DEBUG"] = config.FLASK_DEBUG
    app.config["FLASK_HOST"] = config.FLASK_HOST
    app.config["CELERY_BROKER_URL"] = config.CELERY_BROKER_URL
    app.config["CELERY_BACKEND_URL"] = config.CELERY_BACKEND_URL

    # celery setup
    celery_app = create_celery_app_for_flask(app)
    app.celery = celery_app

    # endpoints
    api = Blueprint("api", __name__)
    api.register_blueprint(generate_blueprint)
    api.register_blueprint(chat_blueprint)
    api.register_blueprint(embeddings_blueprint)
    app.register_blueprint(api)

    app.run(
        debug=app.config["FLASK_DEBUG"],
        host=app.config["FLASK_HOST"],
        port=app.config["FLASK_RUN_PORT"],
    )

    return app



"""
Generate a chat completion
API implementation of `POST /api/chat` endpoint, handling completion orchestration, as replica of the same Ollama server endpoint.
Ollama endpoint reference: https://github.com/ollama/ollama/blob/main/docs/api.md#generate-a-chat-completion
"""

import time
from flask import Blueprint, request
from oshepherd.api.utils import streamify_json
from oshepherd.api.chat.models import ChatRequest

chat_blueprint = Blueprint("chat", __name__, url_prefix="/api/chat")


@chat_blueprint.route("/", methods=["POST"])
def chat():
    from oshepherd.worker.tasks import exec_completion

    print(f" # request.json {request.json}")
    chat_request = ChatRequest(**{"payload": request.json})

    # req as json string ready to be sent through broker
    chat_request_json_str = chat_request.model_dump_json()
    print(f" # chat request {chat_request_json_str}")

    # queue request to remote ollama api server
    task = exec_completion.delay(chat_request_json_str)
    while not task.ready():
        print(" > waiting for response...")
        time.sleep(1)
    ollama_res = task.get(timeout=1)

    status = 200
    if ollama_res.get("error"):
        ollama_res = {
            "error": "Internal Server Error",
            "message": f"error executing completion: {ollama_res['error']['message']}",
        }
        status = 500

    print(f" $ ollama response {status}: {ollama_res}")

    return streamify_json(ollama_res, status)



import os
from dotenv import load_dotenv
from pydantic import BaseModel, ValidationError
from typing import Optional


def load_and_validate_env(file_path: str, Config: BaseModel) -> Optional[BaseModel]:
    # Load .env file into environment variables
    load_dotenv(file_path)

    # Create a dictionary of environment variables
    env_vars = {
        key: os.getenv(key) for key in Config.__annotations__.keys() if os.getenv(key)
    }

    # Validate environment variables against ApiConfig model
    try:
        config = Config(**env_vars)
    except ValidationError as e:
        print(f"Invalid configuration: {e}")
        return None

    return config



import json
from flask import Response, stream_with_context


def streamify_json(json_data, status=200):
    def json_stream(data):
        yield json.dumps(data)

    return Response(
        stream_with_context(json_stream(json_data)),
        status=status,
        mimetype="application/json",
    )



from pydantic import BaseModel
from typing import Optional, List, Literal
from datetime import datetime


class ChatMessage(BaseModel):
    role: Literal["user", "assistant", "system"]
    content: str
    # TODO add support for images
    # images: NotRequired[Sequence[Any]]


class ChatRequestPayload(BaseModel):
    model: str
    messages: Optional[List[ChatMessage]] = None
    format: Optional[str] = ""
    options: Optional[dict] = {}
    stream: Optional[bool] = False
    keep_alive: Optional[str] = None


class ChatRequest(BaseModel):
    type: str = "chat"
    payload: ChatRequestPayload


class ChatResponse(BaseModel):
    model: str
    created_at: datetime
    message: ChatMessage
    done: bool
    total_duration: int
    load_duration: int
    prompt_eval_count: Optional[int] = None
    prompt_eval_duration: int
    eval_count: int
    eval_duration: int



from celery import Celery
from oshepherd.worker.config import WorkerConfig

TASKS_MODULE = "oshepherd.worker.tasks"

celery_app = None


def create_celery_app(config: WorkerConfig):
    """
    Create a Celery app instance given a certain configuration.
    Refs:
        * https://docs.celeryq.dev/en/stable/getting-started/next-steps.html#using-celery-in-your-application
        * https://docs.celeryq.dev/en/stable/getting-started/first-steps-with-celery.html#celerytut-keeping-results
    """
    global celery_app

    celery_app = Celery(
        TASKS_MODULE, broker=config.CELERY_BROKER_URL, backend=config.CELERY_BACKEND_URL
    )
    celery_app.conf.update(
        result_expires=config.RESULTS_EXPIRES,
        broker_transport_options=config.BROKER_TRANSPORT_OPTIONS,
    )
    celery_app.autodiscover_tasks([TASKS_MODULE])

    return celery_app


def create_celery_app_for_flask(flask_app):
    config = WorkerConfig(
        CELERY_BROKER_URL=flask_app.config["CELERY_BROKER_URL"],
        CELERY_BACKEND_URL=flask_app.config["CELERY_BACKEND_URL"],
        RESULTS_EXPIRES=flask_app.config.get(
            "RESULTS_EXPIRES", WorkerConfig.__fields__["RESULTS_EXPIRES"].default
        ),
    )

    celery_app = create_celery_app(config)

    return celery_app



from pydantic import BaseModel
from typing import Optional


class ApiConfig(BaseModel):
    FLASK_PROJECT_NAME: Optional[str] = "oshepherd_api"
    FLASK_ENV: Optional[str] = "development"
    FLASK_DEBUG: Optional[bool] = True
    FLASK_RUN_PORT: Optional[int] = 5001
    FLASK_HOST: Optional[str] = "0.0.0.0"
    CELERY_BROKER_URL: str
    CELERY_BACKEND_URL: str
